package kw4h;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * Calculator class - to calculate
 * @author Kw4h
 *
 */
public class KwCalc 
{
	// Objects needed
	private Map shortestDistances = new HashMap();
	private Set settledSystems = new HashSet();
	private Set unsettledSystems = new HashSet();
	private Map predecessors = new HashMap();
	private int avoidance = 0;
	private Hashtable jumps;
	private Vector avoidSystems = new Vector();
	
	private Logger logger = Logger.getLogger(this.getClass());
	
	/**
	 * Calculates distance between two 3d points
	 * @param x1
	 * @param y1
	 * @param z1
	 * @param x2
	 * @param y2
	 * @param z2
	 * @return
	 */
	public static double calcDistance(double[] pos1, double[] pos2)
	{
		double x1, y1, z1, x2, y2, z2;
		x1 = pos1[0];
		y1 = pos1[1];
		z1 = pos1[2];
		x2 = pos2[0];
		y2 = pos2[1];
		z2 = pos2[2];
		double distance = Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow((y2 - y1), 2) + Math.pow((z2 - z1), 2));
		
		return distance;
	}
	
	public KwCalc()
	{
		logger.info("Creating KwCalc instance");
	}
	
	/**
	 * 
	 * @param xCam
	 * @param yCam
	 * @param zCam
	 * @param xTarget
	 * @param yTarget
	 * @param zTarget
	 * @return
	 */
	public static double calcViewAngle(double xCam, double yCam, double zCam, double xTarget, double yTarget, double zTarget)
	{
		// first run
		double l = Math.sqrt(Math.pow(xTarget - xCam, 2) + Math.pow(yTarget - yCam, 2));
		
		double angle = Math.toDegrees(Math.tan(l / (zTarget - zCam)));
		
		return angle;
	}
	
	/**
	 * 
	 * @param start
	 */
	private void doInit(SolarSystem start)
	{
		// clear everything
		this.settledSystems.clear();
		this.unsettledSystems.clear();
		this.shortestDistances.clear();
		this.predecessors.clear();
		
		// now add the source system
		this.setShortestDistance(start, 0);
		this.unsettledSystems.add(start);
	}
	
	/**
	 * 
	 * @param sys
	 * @param distance
	 */
	private void setShortestDistance(SolarSystem sys, int distance)
	{
		// remove system from the unsettled nodes
		this.unsettledSystems.remove(sys);
		
		// set shortest distance
		this.shortestDistances.put(sys, new Integer(distance));
		this.unsettledSystems.add(sys);
	}
	
	/**
	 * 
	 * @return
	 */
	private SolarSystem extractMin()
	{
		// just get the next system. Distance not yet implemented
		if (this.unsettledSystems.isEmpty())
		{
			return null;
		}
		
		// get the first system
		SolarSystem next = (SolarSystem) this.unsettledSystems.iterator().next();
		this.unsettledSystems.remove(next);
		return next;
	}
	
	/**
	 * 
	 * @param u
	 */
	private void markSettled(SolarSystem u)
	{
		this.settledSystems.add(u);
	}
	
	/**
	 * 
	 * @param v
	 * @return
	 */
	private boolean isSettled(SolarSystem v)
	{
		return this.settledSystems.contains(v);
	}
	
	/**
	 * 
	 * @param sys
	 * @return
	 */
	private int getShortestDistance(SolarSystem sys)
	{
		// get the shortest distance
		Integer d = (Integer) this.shortestDistances.get(sys);
		if (d == null)
		{
			d = Integer.MAX_VALUE;
		}
		
		return d;
	}
	
	/**
	 * 
	 * @param a
	 * @param b
	 */
	private void setPredecessor(SolarSystem a, SolarSystem b)
	{
		// remove older predecessor
		this.predecessors.remove(a);
		
		this.predecessors.put(a, b);
	}
	
	/**
	 * 
	 * @param u
	 */
	private void relaxNeighbours(SolarSystem u)
	{
		// get all neighbours of the system
		Vector neighbours = null;
		neighbours = (Vector) this.jumps.get(u);		
		
		// loop through them
		if (neighbours != null)
		{
			for(Iterator i = neighbours.iterator(); i.hasNext();)
			{
				SolarSystem v = (SolarSystem) i.next();
				
				// are we avoiding this system?
				if (!this.avoidSystems.contains(v.name))
				{
					// get security status
					double sec = v.security;
					if (sec < 0) sec = 0.0;
					
					// line quality
					int quality = 5;
					
					// preferring highsec
					if (this.avoidance == 1)
					{
						if (sec >= 0.5) quality = 1;
					}
					// preferring lowsec
					else if (this.avoidance == 2)
					{
						if (sec < 0.5) quality = 1;
					}
					
					// shorter one exists - better quality (lower quality / shorter quality)
					if ((this.getShortestDistance(u) + quality) < this.getShortestDistance(v))
					{
						//System.out.println("=========== " + u.name + " ===========");
						//System.out.println(u.name + ": " + (this.getShortestDistance(u) + quality) + " shorter than " + v.name + ": " + this.getShortestDistance(v));
						this.setShortestDistance(v, (this.getShortestDistance(u) + quality));
						this.setPredecessor(v, u);
					}
				}
			}
		}
		else
		{
			System.out.println ("No jumps for: " + u.name);
		}
	}
	
	/**
	 * 
	 * @param jumps
	 * @param start
	 * @param destination
	 * @param avoidance
	 * @return
	 */
	public Vector calcJumps(Hashtable jumps, SolarSystem start, SolarSystem destination, int avoidance, Vector avoidSystems)
	{
		// many thanks to: http://renaud.waldura.com/doc/java/dijkstra/
		// from	30004469 (ZS-2LT)
		// to 	30004450 (AZN-D2)
		// Jumps {(SolarSystem) from, Vector destination {(SolarSystem to}}
		
		// avoidance
		// 0 = no avoidance
		// 1 = Prefer Highsec
		// 2 = Prefer Lowsec
		
		logger.info("Starting dijkstra's algorithm");
		
		this.jumps = jumps;
		this.avoidance = avoidance;
		this.avoidSystems = avoidSystems;
		
		// initialise
		this.doInit(start);
		
		// the current system
		SolarSystem u;
		
		// do the loop
		while((u = this.extractMin()) != null)
		{
			// destination reached - stop
			/*if (u == destination)
			{
				break;
			}*/
			
			this.relaxNeighbours(u);
		}
		
		// result vector
		Vector jumpers = new Vector();
		jumpers.add(destination);
		
		SolarSystem echo;
		echo = (SolarSystem) this.predecessors.get(destination);
		jumpers.add(echo);
		
		while ((echo = (SolarSystem) this.predecessors.get(echo)) != null)
		{
			jumpers.add(echo);
		}
		
		logger.info("Ending dijkstra's algorithm");
		return jumpers;
	}

}
